package ca.yorku.cse.mack.h4touch;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Locale;
import java.util.Random;
import ca.yorku.cse.mack.h4touch.H4Keyboard.H4Listener;
import android.app.Activity;
import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.res.Configuration;
import android.media.MediaPlayer;
import android.media.MediaScannerConnection;
import android.os.Bundle;
import android.os.Environment;
import android.os.Vibrator;
import android.util.DisplayMetrics;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.widget.EditText;
import android.widget.TextView;

/**
 * H4Writer - Android implementation of the H4Writer text entry method.
 * <p>
 * 
 * Includes a setup dialog to choose parameters for experimental testing.
 * <p>
 * 
 * @author Scott MacKenzie
 * 
 */

public class H4TouchActivity extends Activity implements H4Listener
{
	final String WORKING_DIRECTORY = "/H4TouchData/";
	final String SD2_HEADER = "Participant,Session,Block,Group,Condition,Key_style,"
			+ "Min_keystrokes,Actual_keystrokes,Presented_char,Transcribed_char,Time(s),MSD,"
			+ "Entry_speed(wpm),Error_rate(%),KSPC,Efficiency\n";

	final int KEYSTROKES_FOR_ENTER = 2; // adjustment needed for output data (keystrokes for ENTER not counted)
	final long PULSE_DURATION = 5;
	final int MARGIN_LEFT = 5; // should be consistent with values in main.xml

	// setup parameters
	String participantCode, sessionCode, blockCode, groupCode, conditionCode;
	String[] phrases;
	boolean vibrotactileFeedback, auditoryFeedback;
	boolean showText;
	boolean lowercaseOnly, showHuffman;
	// end setup parameters

	boolean phraseStarted;
	boolean newPhrase = true;
	boolean done = false;
	Vibrator vib;
	MediaPlayer enter, tick, miss;
	BufferedWriter sd1, sd2;
	File f1, f2;
	EditText presentedTextField, transcribedTextField;
	H4Keyboard h4k;
	String presentedText, transcribedText;
	Random r = new Random();
	int key;
	int keystrokeCount, phraseCount, phrasesPerBlock;
	long timeStampStart, timeStampEnd;
	int timeForPhrase;
	int keyStyle;
	String keyStyleString;

	@Override
	public void onCreate(Bundle savedInstanceState)
	{
		super.onCreate(savedInstanceState);
		setContentView(R.layout.main);

		Bundle b = getIntent().getExtras();
		participantCode = b.getString("participantCode");
		sessionCode = b.getString("sessionCode");
		blockCode = b.getString("blockCode");
		groupCode = b.getString("groupCode");
		conditionCode = b.getString("conditionCode");
		phrasesPerBlock = b.getInt("phrasesPerBlock");
		vibrotactileFeedback = b.getBoolean("vibrotactileFeedback");
		auditoryFeedback = b.getBoolean("auditoryFeedback");
		showText = b.getBoolean("showText");
		lowercaseOnly = b.getBoolean("lowercaseOnly");
		showHuffman = b.getBoolean("showHuffmanCodes");

		// special fiddling required for phrases array
		if (b.getString("phrasesFile").equals("phrases2"))
			phrases = getResources().getStringArray(R.array.phrases2);
		else if (b.getString("phrasesFile").equals("quickbrownfox"))
			phrases = getResources().getStringArray(R.array.quickbrownfox);
		else if (b.getString("phrasesFile").equals("phrases100"))
			phrases = getResources().getStringArray(R.array.phrases100);
		else if (b.getString("phrasesFile").equals("alphabet"))
			phrases = getResources().getStringArray(R.array.alphabet);

		// special fiddling required for key style
		keyStyle = b.getString("keyStyle").equals("x") ? H4Keyboard.TRIANGLE : H4Keyboard.SQUARE;
		if (keyStyle == H4Keyboard.TRIANGLE)
			keyStyleString = "TR";
		else
			keyStyleString = "SQ";

		// ===================
		// File initialization
		// ===================

		// make a working directory to store sd1 and sd2 data files
		File dataDirectory = new File(Environment.getExternalStorageDirectory() + WORKING_DIRECTORY);
		if (!dataDirectory.exists() && !dataDirectory.mkdirs())
		{
			super.onDestroy(); // cleanup
			this.finish(); // terminate
		}

		// base filename for output data files (find first available block code and use in filename)
		int blockNumber = 1;
		blockCode = "B01";
		String base = "H4Touch-" + participantCode + "-" + sessionCode + "-" + blockCode + "-" + groupCode + "-"
				+ conditionCode + "-" + keyStyleString;

		f1 = new File(dataDirectory, base + ".sd1");
		f2 = new File(dataDirectory, base + ".sd2");

		// check with user before overwriting
		while (f1.exists() || f2.exists())
		{
			++blockNumber;
			blockCode = blockNumber < 10 ? "B0" + blockNumber : "B" + blockNumber;
			base = "H4Touch-" + participantCode + "-" + sessionCode + "-" + blockCode + "-" + groupCode + "-"
					+ conditionCode + "-" + keyStyleString;

			f1 = new File(dataDirectory, base + ".sd1");
			f2 = new File(dataDirectory, base + ".sd2");
		}

		try
		{
			sd1 = new BufferedWriter(new FileWriter(f1));
			sd2 = new BufferedWriter(new FileWriter(f2));

			// output header in sd2 file
			sd2.write(SD2_HEADER, 0, SD2_HEADER.length());
			sd2.flush();
		} catch (IOException e)
		{
			super.onDestroy();
			this.finish();
		}
		// end file initialization

		presentedTextField = (EditText)findViewById(R.id.presented);
		transcribedTextField = (EditText)findViewById(R.id.transcribed);

		// prevent system's IME soft keyboard from popping up
		getWindow().setSoftInputMode(WindowManager.LayoutParams.SOFT_INPUT_STATE_ALWAYS_HIDDEN);

		h4k = (H4Keyboard)findViewById(R.id.h4keys);
		h4k.setH4Listener(this);
		h4k.showHuffmanCodes = showHuffman;
		h4k.keyStyle = keyStyle;

		// determine display width and height
		DisplayMetrics dm = new DisplayMetrics();
		getWindowManager().getDefaultDisplay().getMetrics(dm);
		float scalingFactor = dm.density;
		int width = dm.widthPixels - (int) (2 * MARGIN_LEFT * scalingFactor + 0.5f);
		h4k.setKeyboardSize(width, dm.heightPixels, scalingFactor);

		// initialize vibrotactile and auditory feedback objects
		vib = (Vibrator)getSystemService(Context.VIBRATOR_SERVICE);
		tick = MediaPlayer.create(getApplicationContext(), R.raw.click1);
		miss = MediaPlayer.create(getApplicationContext(), R.raw.miss);
		enter = MediaPlayer.create(getApplicationContext(), R.raw.blip1);

		phraseCount = 0;
		prepareForPhrase();
	}

	public void onH4Code(H4Keyboard.H4Event h4e)
	{
		String symbol = h4e.symbol;
		long timeStamp = h4e.timeStamp;

		if (!phraseStarted)
		{
			phraseStarted = true;
			timeStampStart = h4e.timeStamp;
		}

		if (!symbol.equals(H4Keyboard.ENTER))
		{
			// first, update the transcribed text field, as appropriate, and output the appropriate feedback
			if (symbol.equals(H4Keyboard.SPACE))
			{
				transcribedText += " ";
				if (vibrotactileFeedback)
					vib.vibrate(PULSE_DURATION);
				if (auditoryFeedback)
					tick.start();
			} else if (symbol.equals(H4Keyboard.BACKSPACE))
			{
				if (transcribedText.length() >= 1)
				{
					transcribedText = transcribedText.substring(0, transcribedText.length() - 1);
					if (vibrotactileFeedback)
						vib.vibrate(PULSE_DURATION);
					if (auditoryFeedback)
						tick.start();
				} else if (auditoryFeedback)
					miss.start();
			} else
			{
				transcribedText += symbol;
				if (vibrotactileFeedback)
					vib.vibrate(PULSE_DURATION);
				if (auditoryFeedback)
					tick.start();
			}
			transcribedTextField.setText(transcribedText);
			transcribedTextField.setSelection(transcribedText.length()); // move cursor to end
		} else
		// ENTER (end of phrase)
		{
			doEndOfPhrase(timeStamp);
		}
	}

	// count keystrokes
	public void onH4Keystroke()
	{
		++keystrokeCount;

		if (done)
			doDone();
	}

	private void doEndOfPhrase(long timeStampEnd)
	{
		if (vibrotactileFeedback)
			vib.vibrate(PULSE_DURATION);
		if (auditoryFeedback)
			enter.start();

		newPhrase = true;
		StringBuilder resultsString = new StringBuilder(500);
		resultsString.append("Thank you!\n");

		// don't count the keystrokes for ENTER
		keystrokeCount -= KEYSTROKES_FOR_ENTER;
		timeForPhrase = (int) (timeStampEnd - timeStampStart);

		String s1 = presentedText;
		String s2 = transcribedText.trim();
		if (lowercaseOnly)
		{
			s1 = s1.toLowerCase(Locale.US);
			s2 = s2.toLowerCase(Locale.US);
		}

		resultsString.append("Presented...\n   " + s1 + "\n");
		resultsString.append("Transcribed...\n   " + s2 + "\n");

		StringBuilder sd2Stuff = new StringBuilder(500);
		sd2Stuff.append(participantCode + "," + // participant code
				sessionCode + "," + // session code
				blockCode + "," + // block code
				groupCode + "," + // group code
				conditionCode + "," + // condition code
				keyStyleString + ","); // key style (TR = triangular, SQ = square)

		// MINIMUM NUMBER OF KEYSTROKES
		sd2Stuff.append(minimumKeystrokes(presentedText) + ",");

		// ACTUAL NUMBER OF KEYSTROKES
		sd2Stuff.append(keystrokeCount + ",");

		// PRESENTED CHARACTERS (PHRASE LENGTH)
		sd2Stuff.append(s1.length() + ",");

		// NUMBER OF CHARACTERS OF TEXT PRODUCED
		sd2Stuff.append(s2.length() + ",");

		// TIME (SECONDS)
		float d = trim(timeForPhrase / 1000.0f, 2);
		sd2Stuff.append(d + ",");

		// TIME (MINUTES)
		d = trim(timeForPhrase / 1000.0f / 60.0f, 3);

		// MSD
		MSD s1s2 = new MSD(s1.toLowerCase(Locale.US).trim(), s2.toLowerCase(Locale.US).trim());
		int msd = s1s2.getMSD();
		sd2Stuff.append(msd + ",");

		// ENTRY SPEED (WPM)
		d = trim(wpm(s2, timeForPhrase), 2);
		sd2Stuff.append(d + ",");
		resultsString.append("Entry speed = " + d + " wpm\n");

		// ERROR RATE (%)
		d = (float)s1s2.getErrorRateNew();
		d = trim(d, 2);
		sd2Stuff.append(d + ",");
		resultsString.append("Error rate = " + d + "%\n");

		// KSPC
		float kspc = (float)keystrokeCount / s2.length();
		sd2Stuff.append(kspc + ",");

		// EFFICIENCY
		float kspcmin = (float)minimumKeystrokes(presentedText) / s1.length();
		float efficiency = (kspcmin / kspc) * 100.0f;
		resultsString.append("Extra taps = " + (keystrokeCount - minimumKeystrokes(presentedText)));
		sd2Stuff.append(efficiency + ",");

		// present results to user
		showResultsDialog(resultsString.toString());

		// build sd1 data (and compute mean distance from center for sd2 file)
		StringBuilder sd1Stuff = new StringBuilder(500);
		sd1Stuff.append(presentedText + "\n" + s2 + "\n");

		// dump data to sd1 and sd2 files
		try
		{
			sd1.write(sd1Stuff.toString(), 0, sd1Stuff.length());
			sd1.flush();
			sd2.write(sd2Stuff.toString(), 0, sd2Stuff.length());
			sd2.flush();
		} catch (IOException e)
		{
			// Log.d("MYDEBUG", "ERROR WRITING TO DATA FILE! e = " + e);
			this.finish();
		}

		// terminate if this was the last phrase in the block
		if (++phraseCount == phrasesPerBlock)
			done = true; // app will terminate on next touch (display results for last phrase)
		else
			prepareForPhrase();
	}

	private void showResultsDialog(String text)
	{
		LayoutInflater inflater = (LayoutInflater)this.getSystemService(LAYOUT_INFLATER_SERVICE);
		View layout = inflater.inflate(R.layout.results_dialog, (ViewGroup)findViewById(R.id.results_layout));

		// Set text
		TextView results = (TextView)layout.findViewById(R.id.resultsArea);
		results.setText(text);

		// Initialize the dialog
		AlertDialog.Builder parameters = new AlertDialog.Builder(this);
		parameters.setView(layout).setCancelable(false).setNeutralButton("OK", new DialogInterface.OnClickListener()
		{
			public void onClick(DialogInterface dialog, int id)
			{
				dialog.cancel(); // close this dialog
			}
		}).show();
	}

	private void doDone()
	{
		try
		{
			sd1.close();
			sd2.close();
			
			// Make the saved data files visible in Windows Explorer
			// There seems to be bug doing this with Android 4.4.  I'm using the following
			// code, instead of sendBroadcast.  See...
			// http://code.google.com/p/android/issues/detail?id=38282
			MediaScannerConnection.scanFile(this, new String[] { f1.getAbsolutePath(),
					f2.getAbsolutePath() }, null, null);
		} catch (IOException e)
		{
			// Log.i("MYDEBUG", "ERROR CLOSING DATA FILES! e = " + e);
		}
		super.onDestroy();
		this.finish();
	}

	/**
	 * Compute text entry speed in words per minute (wpm).
	 * 
	 * @param text
	 *            a phrase of text
	 * @param msTime
	 *            time to enter the phrase in milliseconds
	 * @return entry speed in words per minute (wpm) or -1 if time is <= 0
	 */
	public static float wpm(String text, long msTime)
	{
		float speed = text.length();
		if (msTime > 0)
			return speed / (msTime / 1000.0f) * 12.0f; // Note: 60 / 5 = 12
		else
			return -1f;
	}

	/**
	 * Compute the minimum number of keystrokes to enter a phrase.
	 * <p>
	 * 
	 * NOTE: The value returned excludes the keystrokes for a terminator, such as ENTER.
	 * <p>
	 * 
	 * @param phrase
	 *            the phrase (e.g., "the quick brown fox")
	 * @return minimum number of keystrokes to enter the phrase, or -1 if phrase length is zero
	 */
	private int minimumKeystrokes(String phrase)
	{
		int count = 0;
		for (int i = 0; i < phrase.length(); ++i)
		{
			Character c = Character.valueOf(phrase.charAt(i));
			String s = "" + c;

			// some pre-amble adjustments for SPACE, uppercase, and non-letters
			if (Character.isSpaceChar(c))
				s = H4Keyboard.SPACE; // re-map, if SPACE

			if (Character.isUpperCase(c)) // add transition to caps
			{
				count += sizeOfCode(H4Keyboard.SHIFT, H4Keyboard.LETTER_CODES);
				s = s.toLowerCase(Locale.US);
			}

			if (Character.isLetter(c) || Character.isSpaceChar(c) || s.equals(".") || s.equals(","))
			{
				count += sizeOfCode(s, H4Keyboard.LETTER_CODES);
			} else
			{
				count += sizeOfCode(H4Keyboard.SYMBOL, H4Keyboard.LETTER_CODES);
				count += sizeOfCode(s, H4Keyboard.SYMBOL_CODES);
			}
		}
		return count > 0 ? count : -1;
	}

	/**
	 * Compute the size of a huffman code for a given symbol.
	 * 
	 * @param symbol
	 * @param huffmanCodes
	 * @return the size of the code, or -1 if the code is not found
	 */
	private int sizeOfCode(String symbol, String[][] huffmanCodes)
	{
		int count = 0;
		for (int i = 0; i < huffmanCodes.length; ++i)
		{
			if (symbol.equals(huffmanCodes[i][0]))
			{
				count += huffmanCodes[i][1].length();
				break;
			}
		}
		return count > 0 ? count : -1;
	}

	void prepareForPhrase()
	{
		String phrase = phrases[r.nextInt(phrases.length)];
		presentedText = lowercaseOnly ? phrase.toLowerCase(Locale.US) : phrase;
		presentedTextField.setText(presentedText);
		transcribedText = "";
		transcribedTextField.setText(transcribedText);
		phraseStarted = false;
		keystrokeCount = 0;
		return;
	}

	public void onConfigurationChanged(Configuration newConfig)
	{
		// because the manifest file includes "android:configChanges="orientation"", orientation changes are handled
		// here
		super.onConfigurationChanged(newConfig);

		// determine display width and height
		DisplayMetrics dm = new DisplayMetrics();
		getWindowManager().getDefaultDisplay().getMetrics(dm);
		float scalingFactor = dm.density;
		int width = dm.widthPixels - (int) (2 * MARGIN_LEFT * scalingFactor + 0.5f);
		h4k.setKeyboardSize(width, dm.heightPixels, scalingFactor);
	}

	public void onSaveInstanceState(Bundle savedInstanceState)
	{
		super.onSaveInstanceState(savedInstanceState);
	}

	public void onRestoreInstanceState(Bundle savedInstanceState)
	{
		super.onRestoreInstanceState(savedInstanceState);
	}

	// trim and round a float to the specified number of decimal places
	private float trim(float f, int decimalPlaces)
	{
		return (int) (f * 10 * decimalPlaces + 0.5f) / (float) (10 * decimalPlaces);
	}
}